package com.flow;

import android.app.Activity;
import android.content.Context;
import android.graphics.Point;
import android.provider.Settings;
import android.util.SparseArray;
import android.view.View;
import android.view.ViewGroup;
import android.view.ViewParent;
import android.widget.ScrollView;
import android.widget.Toast;

import androidx.annotation.NonNull;
import androidx.annotation.Nullable;

import com.facebook.infer.annotation.Assertions;
import com.facebook.react.bridge.Arguments;
import com.facebook.react.bridge.Callback;
import com.facebook.react.bridge.ReactApplicationContext;
import com.facebook.react.bridge.ReactContext;
import com.facebook.react.bridge.ReactContextBaseJavaModule;
import com.facebook.react.bridge.ReactMethod;
import com.facebook.react.bridge.ReadableArray;
import com.facebook.react.bridge.ReadableMap;
import com.facebook.react.bridge.WritableMap;
import com.facebook.react.bridge.WritableNativeMap;
import com.facebook.react.modules.core.DeviceEventManagerModule;
import com.facebook.react.uimanager.PixelUtil;
import com.facebook.react.uimanager.RootView;
import com.facebook.react.uimanager.RootViewUtil;
import com.facebook.react.views.scroll.ReactScrollView;

import org.json.JSONException;
import org.json.JSONObject;

import java.util.ArrayList;
import java.util.LinkedHashMap;
import java.util.List;
import java.util.Map;

import retrofit2.Call;
import retrofit2.Response;

public class FlowdashNativeModule extends ReactContextBaseJavaModule {

    private static final String TAG = "FlowdashNativeModule";
    private static APIInterface api;
    private String deviceId = Settings.Secure.getString(getReactApplicationContext().getContentResolver(),
            Settings.Secure.ANDROID_ID);
    private Map<String, FetchWalkthroughResponseInternal> capturedViews = new LinkedHashMap<>();
    private static String ownerId;
    private boolean includeEditor;
    private static String apiKey;
    private static Context context;
    private String previousSavedScreenId;

    FlowdashNativeModule(@NonNull ReactApplicationContext reactContext, String _ownerId, String ownerKey, boolean includeEditor) {
        super(reactContext);
        api = APIClient.getClient().create(APIInterface.class);
        apiKey = ownerKey;
        ownerId = _ownerId;
        context = reactContext.getApplicationContext();
        this.includeEditor = includeEditor;
    }

    @NonNull
    @Override
    public String getName() {
        return "Flowdash";
    }

    @Override
    public void initialize() {
        Logger.i(TAG, "FlowdashNativeModule initialized. ");
        fetchWalkthrough(ownerId);
    }

    @ReactMethod(isBlockingSynchronousMethod = true)
    public boolean isEditorEnabled() {
        return includeEditor;
    }

    @ReactMethod
    public void showToastWithText(String message) {
        Toast.makeText(getReactApplicationContext(), message, Toast.LENGTH_LONG).show();
    }

    @ReactMethod
    public void saveView(final String screenId, ReadableArray views) {
        final List<FlowDashCapturedViewIndependant> viewsArray = new ArrayList<>();
        List<FlowDashCaptureViewModel> models = new ArrayList<>();
        try {
            models = FlowDashUtils.convertJsonArrayToList(FlowDashUtils.convertArrayToJsonArray(views), FlowDashCaptureViewModel.class);
        } catch (JSONException e) {
            Logger.w(TAG, "Error parsing views.", e);
        }
        final Activity activity = getCurrentActivity();
        final View rootView = activity.getWindow().getDecorView().getRootView();
        if (rootView == null) {
            Logger.w(TAG, "RootView is null.");
            return;
        }
        ViewHierarchy parsedViews = new ViewHierarchy(rootView, new SparseArray());
        try {
            parsedViews.parseViewHierarchy();
        } catch (Throwable throwable) {
            Logger.e(TAG, "saveView: Error parsing view hierarchy", throwable);
        }

        for (FlowDashCaptureViewModel viewModel : models) {
            Logger.d(TAG, "saveView: Native view id :" + viewModel.getNativeViewTag());

            View requiredView = rootView.findViewById(viewModel.getNativeViewTag());

            View reactRootView = (View) RootViewUtil.getRootView(requiredView);

            try {

                parsedViews.setCurrentView(requiredView);
                parsedViews.toXMLString();
                viewModel.setViewIndex(parsedViews.getCurrentKey());

                Logger.d(TAG, "Current View index:" + viewModel.getViewIndex());

                int[] location = parsedViews.getViewRectByCurrentView(viewModel.getViewIndex());

                int[] container = parsedViews.customMeasure(reactRootView);

                FlowDashCapturedViewIndependant viewModel1 = new FlowDashCapturedViewIndependant(viewModel);

                viewModel1.setLeftIndependant(PixelUtil.toDIPFromPixel(location[0] - container[0]));
                viewModel1.setTopIndependant(PixelUtil.toDIPFromPixel(location[1] - container[1]));
                viewModel1.setWidthIndependant(PixelUtil.toDIPFromPixel(location[2]));
                viewModel1.setHeightIndependant(PixelUtil.toDIPFromPixel(location[3]));

                viewsArray.add(viewModel1);

            } catch (Throwable t) {
                Logger.e(TAG, "saveView: Error", t);
            }
        }
        List<FlowDashCaptureViewModel> steps = new ArrayList<>();
        for (FlowDashCapturedViewIndependant view : viewsArray) {
            steps.add(new FlowDashCaptureViewModel(view));
        }
        capturedViews.put(screenId, new FetchWalkthroughResponseInternal(screenId, steps));
        previousSavedScreenId = screenId;
        Logger.d(TAG, "saveView: Saved successfully." + capturedViews.values());
    }


    @ReactMethod
    public void publishWalkthrough() {
        Logger.d(TAG, "publishWalkthrough: Trying to save walkthroughs");
        saveWalkthrough(ownerId);
    }

    @ReactMethod
    public void getCapturedViews(String screenId, int stepNumber, Callback successCallback) throws JSONException {
        WritableMap resultView = new WritableNativeMap();
        WritableMap nextScreen = new WritableNativeMap();

        List<FlowDashCapturedViewIndependant> views = new ArrayList<>();
        String nextScreenId = "";
        String type = "default";
        Map<String, Object> nextProps;
        FetchWalkthroughResponseInternal temp = capturedViews.get(screenId);
        if (temp != null) {
            for (FlowDashCaptureViewModel step : temp.getSteps()) {
                views.add(new FlowDashCapturedViewIndependant(step));
            }
            if (!views.isEmpty() && stepNumber == views.size() - 1) {
                nextScreenId = temp.getNextScreenId();
                nextProps = temp.getNextProps();
                if (nextScreenId != null && !nextScreenId.equals("")) {
                    nextScreen.putString("nextScreenId", nextScreenId);
                    if (nextProps != null) {
                        nextScreen.putMap("nextProps", FlowDashUtils.convertJsonToMap(new JSONObject(nextProps)));
                    }
                }
            }
            type = temp.getType();
        }
        Activity activity = getCurrentActivity();
        View rootView = activity.getWindow().getDecorView().getRootView();
        if (rootView == null) {
            Logger.w(TAG, "RootView is null.");
            return;
        }
        ViewHierarchy parsedViews = new ViewHierarchy(rootView, new SparseArray());
        try {
            parsedViews.parseViewHierarchy();
        } catch (Throwable throwable) {
            Logger.e(TAG, "getCapturedViews: Error parsing view hierarchy", throwable);
        }

        if (!views.isEmpty() && stepNumber < views.size()) {
            FlowDashCapturedViewIndependant view = views.get(stepNumber);
            if (view != null) {
                Logger.d(TAG, "getCapturedViews: viewIndependant: " + view.toJsonString());
                try {
                    if (view.isDependent()) {

                        int viewIndex = view.getViewIndex();

                        try {
                            Logger.d(TAG, parsedViews.toXMLString());
                        } catch (Throwable throwable) {
                            throwable.printStackTrace();
                        }

                        View requiredView = parsedViews.getCurrentView(viewIndex);

                        if (requiredView == null) {
                            throw new RuntimeException("getCapturedViews: requiredView is null.");
                        }
                        View reactRootView = (View) RootViewUtil.getRootView(requiredView);

                        // Absolute Position in DP for the currentView.
                        int[] location = parsedViews.getViewRectByCurrentView(viewIndex);

                        // Absolute Position of rootView.
                        int[] container = parsedViews.customMeasure(reactRootView);

                        int offsetY = 0;
                        if (view.isScrollToView()) {
                            // ** Handle scroll ***
                            View scrollViewParent = getScrollViewParentOrRoot(requiredView);

                            // Get deepChild Offset
                            Point childOffset = new Point();

                            if (scrollViewParent instanceof ReactScrollView)
                                scrollToView((ScrollView) scrollViewParent, requiredView, childOffset);
                            offsetY = childOffset.y;
                        }

                        // Actual position of currentView.
                        FlowDashCaptureViewModel viewModel = new FlowDashCaptureViewModel(view.getNativeId(),
                                PixelUtil.toDIPFromPixel(location[0] - container[0]),
                                PixelUtil.toDIPFromPixel(location[1] - container[1] - offsetY),
                                PixelUtil.toDIPFromPixel(location[2]),
                                PixelUtil.toDIPFromPixel(location[3]),
                                view.getView(),
                                view.isShowIndex(),
                                view.getViewIndex(),
                                view.getTooltipDirection()
                        );
                        JSONObject jo = new JSONObject(viewModel.toJsonString());
                        resultView.merge(FlowDashUtils.convertJsonToMap(jo));
                        Logger.d(TAG, "getCapturedViews: viewmodel: " + viewModel.toJsonString());
                    } else {
                        FlowDashCaptureViewModel viewModel = new FlowDashCaptureViewModel(view);
                        JSONObject jo = new JSONObject(viewModel.toJsonString());
                        resultView.merge(FlowDashUtils.convertJsonToMap(jo));
                        Logger.d(TAG, "getCapturedViews: viewmodel: " + viewModel.toJsonString());
                    }
                } catch (Exception e) {
                    Logger.e(TAG, "getCapturedViews - error:" + e.getMessage());
                }
            }
        }

        successCallback.invoke(nextScreen, resultView, type);
        WritableMap params = Arguments.createMap();
        params.putString("loading", "false");
        params.putString("screenId", screenId);
        sendEvent(getReactApplicationContext(), "WalkthroughRequest", params);
    }

    @ReactMethod
    public void requestWalkthrough(String screenId, Callback successCallback) {
        WritableMap params = Arguments.createMap();
        params.putString("screenId", screenId);
        params.putString("loading", "true");
        sendEvent(getReactApplicationContext(), "WalkthroughRequest", params);
        successCallback.invoke();
    }

    @ReactMethod
    public void trackCurrentScreen(String screenId, ReadableMap properties) {
        WritableMap params = Arguments.createMap();
        Logger.d(TAG, "trackCurrentScreen: Screen - " + screenId);
        Logger.d(TAG, "trackCurrentScreen: Props - " + properties);
        params.putString("screenId", screenId);
        FetchWalkthroughResponseInternal previousView = capturedViews.get(previousSavedScreenId);
        if (previousView != null && isEditorEnabled()) {
            previousView.setNextScreenId(screenId);
            previousView.setNextProps(properties.toHashMap());
            capturedViews.put(previousSavedScreenId, previousView);
        }
        sendEvent(getReactApplicationContext(), "TrackScreenRequest", params);
    }


    @ReactMethod
    public void recordViewByUser(String screenId, boolean completedStatus) {
        if (capturedViews.get(screenId) != null && !includeEditor) {
            String walkthroughId = capturedViews.get(screenId).getWalkthroughId();
            capturedViews.remove(screenId);
            saveUserView(walkthroughId, deviceId, completedStatus);
        }
    }

    @Override
    public boolean canOverrideExistingModule() {
        return false;
    }


    private void sendEvent(ReactContext reactContext,
                           String eventName,
                           @Nullable WritableMap params) {
        reactContext
                .getJSModule(DeviceEventManagerModule.RCTDeviceEventEmitter.class)
                .emit(eventName, params);
    }

    private void saveWalkthrough(String userId) {
        Logger.d(TAG, "Publishing :" + capturedViews.values());
        Call<Void> call1 = api.saveWalkthrough(apiKey, userId, FlowDashUtils.createDefaultGson().toJson(capturedViews.values()));
        call1.enqueue(new retrofit2.Callback<Void>() {

            @Override
            public void onResponse(@NonNull Call<Void> call, @NonNull Response<Void> response) {
                Void saveResponse = response.body();
                Logger.d(TAG, "saveWalkthrough: Save worked");
            }

            @Override
            public void onFailure(@NonNull Call<Void> call, @NonNull Throwable t) {
                Logger.d(TAG, "saveWalkthrough: Save failed " + t.toString());
                call.cancel();
            }
        });
    }

    private void fetchWalkthrough(String userId) {

        Logger.d(TAG, "fetchWalkthrough: Device:" + deviceId);

        Call<Map<String, String>> call1 = api.fetchWalkthrough(apiKey, userId, deviceId);

        call1.enqueue(new retrofit2.Callback<Map<String, String>>() {

            @Override
            public void onResponse(@NonNull Call<Map<String, String>> call, @NonNull Response<Map<String, String>> response) {
                Map<String, String> fetchResponse = response.body();
                if (fetchResponse != null && !fetchResponse.isEmpty()) {
                    for (Map.Entry<String, String> entry : fetchResponse.entrySet()) {
                        Logger.d(TAG, entry.getKey() + " : " + entry.getValue());
                        List<FetchWalkthroughResponseInternal> list = FetchWalkthroughResponseInternal.listFromJsonString(entry.getValue());
                        for (FetchWalkthroughResponseInternal each : list) {
                            each.setWalkthroughId(entry.getKey());
                            capturedViews.put(each.getScreenId(), each);
                        }
                    }
                    Logger.d(TAG, "fetchWalkthrough: Fetch worked " + fetchResponse.toString());
                } else {
                    Logger.d(TAG, "fetchWalkthrough: Empty Response");
                }
            }

            @Override
            public void onFailure(@NonNull Call<Map<String, String>> call, @NonNull Throwable t) {
                Logger.d(TAG, "fetchWalkthrough: Fetch failed " + t.toString());
                call.cancel();
            }
        });
    }

    private void saveUserView(String walkthroughId, final String deviceId, final boolean completedStatus) {
        Call<Boolean> call1 = api.recordUserView(apiKey, deviceId, walkthroughId, completedStatus);

        call1.enqueue(new retrofit2.Callback<Boolean>() {

            @Override
            public void onResponse(@NonNull Call<Boolean> call, @NonNull Response<Boolean> response) {
                Boolean saveResponse = response.body();
                if (saveResponse != null) {
                    Logger.d(TAG, "saveUserView: Save worked for user" + deviceId);
                }
            }

            @Override
            public void onFailure(@NonNull Call<Boolean> call, @NonNull Throwable t) {
                Logger.d(TAG, "saveUserView: Save failed " + t.toString());
                call.cancel();
            }
        });
    }

    /**
     * Returns the first ScrollView parent of a given view or the RootView
     */
    private View getScrollViewParentOrRoot(View reactView) {
        View current = reactView;
        while (true) {
            if (current instanceof ReactScrollView) {
                return current;
            }
            if (current instanceof RootView) {
                return current;
            }
            ViewParent next = current.getParent();
            if (next == null) {
                return null;
            }
            Assertions.assertCondition(next instanceof View);
            current = (View) next;
        }
    }

    /**
     * Used to scroll to the given view.
     *
     * @param scrollViewParent Parent ScrollView
     * @param view             View to which we need to scroll.
     */
    private void scrollToView(final ScrollView scrollViewParent, final View view, Point childOffset) {
        getDeepChildOffset(scrollViewParent, view.getParent(), view, childOffset);
        // Scroll to child.
        scrollViewParent.smoothScrollTo(0, childOffset.y);
    }

    /**
     * Used to get deep child offset.
     * <p/>
     * 1. We need to scroll to child in scrollview, but the child may not the direct child to scrollview.
     * 2. So to get correct child position to scroll, we need to iterate through all of its parent views till the main parent.
     *
     * @param mainParent        Main Top parent.
     * @param parent            Parent.
     * @param child             Child.
     * @param accumulatedOffset Accumulated Offset.
     */
    private void getDeepChildOffset(final ViewGroup mainParent, final ViewParent parent, final View child, final Point accumulatedOffset) {
        ViewGroup parentGroup = (ViewGroup) parent;
        accumulatedOffset.x += child.getLeft();
        accumulatedOffset.y += child.getTop();
        if (parentGroup.equals(mainParent)) {
            return;
        }
        getDeepChildOffset(mainParent, parentGroup.getParent(), parentGroup, accumulatedOffset);
    }

    public static void logRealTimeErrors(String message) {
        try {
            if (Flowdash.isErrorLoggingEnabled()) {
                Logger.d(TAG, "logRealTimeErrors: Error remote logging is enabled. Attempting to post error log.");
                Call<Void> call1 = api.logError(apiKey, ownerId, FlowDashUtils.createDefaultGson().toJson(new FlowdashError(message, new DeviceInfo(context))));
                call1.enqueue(new retrofit2.Callback<Void>() {

                    @Override
                    public void onResponse(@NonNull Call<Void> call, @NonNull Response<Void> response) {
                        Void saveResponse = response.body();
                        Logger.d(TAG, "logRealTimeErrors: Error logging worked");
                    }

                    @Override
                    public void onFailure(@NonNull Call<Void> call, @NonNull Throwable t) {
                        Logger.d(TAG, "logRealTimeErrors: Error logging failed " + t.toString());
                        call.cancel();
                    }
                });
            }
        } catch (Exception e) {
            Logger.d(TAG, "logRealTimeErrors: Error logging failed " + e.toString());
            //no-op
        }
    }
}
