package com.flow;

import android.graphics.Matrix;
import android.graphics.RectF;
import android.os.SystemClock;
import android.text.TextUtils;
import android.util.SparseArray;
import android.util.Xml;
import android.view.View;
import android.view.ViewGroup;
import android.view.ViewParent;
import android.widget.Adapter;
import android.widget.AdapterView;

import com.facebook.react.uimanager.RootViewUtil;

import org.xmlpull.v1.XmlSerializer;

import java.io.BufferedReader;
import java.io.ByteArrayInputStream;
import java.io.ByteArrayOutputStream;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.InputStream;
import java.io.InputStreamReader;
import java.util.ArrayList;
import java.util.concurrent.Semaphore;
import java.util.regex.Pattern;

import javax.xml.xpath.XPath;
import javax.xml.xpath.XPathFactory;


public final class ViewHierarchy {
    private final Semaphore RESOURCES_GUARD;
    private XmlSerializer serializer;
    private String tmpXmlName;
    private final View root;
    private View currentView;
    private int currentKey;
    private final String TAG = "ViewHierarchy";

    public int getCurrentKey() {
        return currentKey;
    }

    public void setCurrentKey(int currentKey) {
        this.currentKey = currentKey;
    }

    private final SparseArray<View> viewMap;
    private static final XPath xpath = XPathFactory.newInstance().newXPath();
    private static final String NON_XML_CHAR_REPLACEMENT = "?";
    private static final String VIEW_INDEX = "viewIndex";
    private static final String NAMESPACE = "";
    private static final String DEFAULT_VIEW_CLASS_NAME = "android.view.View";
    private static int MAX_TRAVERSAL_DEPTH = 70;
    private static final int MAX_XML_VALUE_LENGTH = 65536;
    private static final String XML_ENCODING = "UTF-8";
    private final RectF mBoundingBox = new RectF();

    public View getCurrentView(int currentViewKey) {
        return this.viewMap.get(currentViewKey);
    }

    public void setCurrentView(View currentView) {
        this.currentView = currentView;
    }


    private static final Pattern XML10_START_TAG_PATTERN = Pattern.compile("^[^_A-Za-zÀ-Öø-˿\u200c-\u200d⁰-\u218fⰀ-\u2fef、-\ud7ff豈-\ufdcfﷰ-�\ud800\udc00-\udbff\udfff]+");
    private static final Pattern XML10_TAG_PATTERN = Pattern.compile("[^\\-._0-9·̀-ͯ‿-⁀A-Za-zÀ-Öø-˿\u200c-\u200d⁰-\u218fⰀ-\u2fef、-\ud7ff豈-\ufdcfﷰ-�\ud800\udc00-\udbff\udfff]");
    private static final ViewHierarchy.Companion Companion = new ViewHierarchy.Companion();

    private void setAttribute(ViewAttributes attrName, Object attrValue) throws IOException {
        if (attrValue != null) {
            String xmlValue = attrValue.toString();
            XmlSerializer serializer = this.serializer;
            if (serializer != null) {
                serializer.attribute("", attrName.toString(), xmlValue);
            }
        }

    }

    private void recordAdapterViewInfo(AdapterView adapterView) throws IOException {
        Adapter adapter = adapterView.getAdapter();
        if (adapter != null) {
            int adapterCount = adapter.getCount();
            ArrayList adapterData = new ArrayList();
            boolean isAdapterTypeSet = false;
            int i = 0;

            for (int count = adapterCount; i < count; ++i) {
                Object adapterItem = adapter.getItem(i);
                if (adapterItem != null) {
                    adapterData.add(adapterItem.toString());
                    if (!isAdapterTypeSet) {
                        this.setAttribute(ViewAttributes.ADAPTER_TYPE, adapterItem.getClass().getSimpleName());
                        isAdapterTypeSet = true;
                    }
                }
            }
            if (!adapterData.isEmpty()) {
                this.setAttribute(ViewAttributes.ADAPTERS, TextUtils.join(",", adapterData));
            }

        }
    }

    private void serializeView(View view, int depth) throws IOException {
        if (view != null) {
            ViewElement viewElement = new ViewElement(view);
            String tagName = Companion.toXmlNodeName(viewElement.getClassName());
            XmlSerializer serializer = this.serializer;
            if (serializer != null) {
                serializer.startTag("", tagName);
            }

            this.setAttribute(ViewAttributes.INDEX, viewElement.getIndex());
            this.setAttribute(ViewAttributes.CLASS, viewElement.getClassName());
            this.setAttribute(ViewAttributes.CONTENT_DESC, viewElement.getContentDescription());
            this.setAttribute(ViewAttributes.CHECKABLE, viewElement.isCheckable());
            this.setAttribute(ViewAttributes.SCROLLABLE, viewElement.isScrollable());
            this.setAttribute(ViewAttributes.LONG_CLICKABLE, viewElement.isLongClickable());
            this.setAttribute(ViewAttributes.PASSWORD, viewElement.isPassword());
            this.setAttribute(ViewAttributes.BOUNDS, viewElement.getBounds().toShortString());
            if (viewElement.getText() != null) {
                ViewText viewText = new ViewText(viewElement.getText().getRawText(), viewElement.getText().isHint());
                this.setAttribute(ViewAttributes.TEXT, viewText.getRawText());
                this.setAttribute(ViewAttributes.HINT, viewText.isHint());
            }

            this.setAttribute(ViewAttributes.RESOURCE_ID, viewElement.getResourceId());
            this.setAttribute(ViewAttributes.VIEW_TAG, viewElement.getViewTag());
            if (view instanceof AdapterView) {
                this.recordAdapterViewInfo((AdapterView) view);
            }

            if (serializer != null) {
                serializer.attribute("", "viewIndex", Integer.toString(this.viewMap.size()));
            }

            this.viewMap.put(this.viewMap.size(), view);

            if (view == this.currentView) {
                setCurrentKey(this.viewMap.size() - 1);
                Logger.d(TAG, "serializeView: Captured view key is " + this.viewMap.size());
            }

            if (depth < MAX_TRAVERSAL_DEPTH) {
                if (view instanceof ViewGroup) {
                    int index = 0;

                    for (int count = ((ViewGroup) view).getChildCount(); index < count; ++index) {
                        this.serializeView(((ViewGroup) view).getChildAt(index), depth + 1);
                    }
                }
            } else {
                Logger.w(TAG, "Skipping traversal of " + view.getClass().getName() + "'s children, since " + "the current depth has reached its maximum allowed value of " + depth);
            }

            if (serializer != null) {
                serializer.endTag("", tagName);
            }

        }
    }

    public void parseViewHierarchy() throws Throwable {
        Throwable lastError = null;
        View rootView = this.root;
        Class[] temp = new Class[]{ByteArrayOutputStream.class, FileOutputStream.class};

        for (int index = 0; index < temp.length; ++index) {
            Class streamType = temp[index];
            this.serializer = Xml.newSerializer();
            this.viewMap.clear();

            try {
                ByteArrayOutputStream outputStream = new ByteArrayOutputStream();

                try {
                    XmlSerializer serializer = this.serializer;
                    if (serializer != null) {
                        serializer.setOutput(outputStream, "UTF-8");
                        serializer.startDocument("UTF-8", true);
                        serializer.setFeature("http://xmlpull.org/v1/doc/features.html#indent-output", true);
                        long startTime = SystemClock.uptimeMillis();
                        this.serializeView(rootView, 0);
                        serializer.endDocument();
                        Logger.d(TAG, "The source XML tree has been fetched in " + (SystemClock.uptimeMillis() - startTime) + "ms " + "using " + streamType.getSimpleName());
                    }
                } catch (OutOfMemoryError ex) {
                    lastError = ex;
                    continue;
                } finally {
                    outputStream.close();
                }

                outputStream.reset();
                return;
            } catch (IOException ex) {
                lastError = (Throwable) ex;
                throw new IOException(lastError);
            }
        }

    }

    private InputStream toStream() throws Throwable {
        Throwable lastError = null;
        View rootView = this.root;
        Class[] temp = new Class[]{ByteArrayOutputStream.class, FileOutputStream.class};

        for (int index = 0; index < temp.length; ++index) {
            Class streamType = temp[index];
            this.serializer = Xml.newSerializer();
            this.viewMap.clear();

            try {
                ByteArrayOutputStream outputStream = new ByteArrayOutputStream();

                try {
                    XmlSerializer serializer = this.serializer;
                    if (serializer != null) {
                        serializer.setOutput(outputStream, "UTF-8");
                        serializer.startDocument("UTF-8", true);
                        serializer.setFeature("http://xmlpull.org/v1/doc/features.html#indent-output", true);
                        long startTime = SystemClock.uptimeMillis();
                        this.serializeView(rootView, 0);
                        serializer.endDocument();
                        Logger.d(TAG, "The source XML tree has been fetched in " + (SystemClock.uptimeMillis() - startTime) + "ms " + "using " + streamType.getSimpleName());
                    }
                } catch (OutOfMemoryError ex) {
                    lastError = ex;
                    continue;
                } finally {
                    outputStream.close();
                }

                return new ByteArrayInputStream(outputStream.toByteArray());
            } catch (IOException ex) {
                lastError = (Throwable) ex;
                throw new IOException(lastError);
            }
        }

        return new ByteArrayInputStream(new byte[0]);

    }

    private final void performCleanup() {
        if (this.tmpXmlName != null) {
            this.tmpXmlName = (String) null;
        }

    }


    public final String toXMLString() throws Throwable {

        try {
            InputStream stream = this.toStream();

            String streamString;
            try {
                StringBuilder sb = new StringBuilder();
                BufferedReader reader = new BufferedReader(new InputStreamReader(stream, "UTF-8"));

                for (String line = reader.readLine(); line != null; line = reader.readLine()) {
                    sb.append(line);
                }

                streamString = sb.toString();
            } catch (Throwable t) {
                throw t;
            }
            return streamString;
        } catch (IOException ex) {
            throw ex;
        }
    }


    public final int[] getViewRectByCurrentView(int currentKey) {
        View view = this.viewMap.get(currentKey);
        return customMeasure(view);
    }

    public int[] customMeasure(View view) {
        int[] location = new int[4];

        View rootView = (View) RootViewUtil.getRootView(view);
        // It is possible that the RootView can't be found because this view is no longer on the screen
        // and has been removed by clipping
        if (rootView == null) {
            Logger.e(TAG, "Root view does not exist");
            return location;
        }

        computeBoundingBox(rootView, location);
        int rootX = location[0];
        int rootY = location[1];
        computeBoundingBox(view, location);
        location[0] -= rootX;
        location[1] -= rootY;
        computeBoundingBox(view, location);
        Logger.d(TAG, "getViewByKey: x - " + location[0] + " y - " + location[1]);
        Logger.d(TAG, "getViewByKey: width - " + location[2] + " y - " + location[3]);
        return location;
    }

    private void computeBoundingBox(View view, int[] outputBuffer) {
        mBoundingBox.set(0, 0, view.getWidth(), view.getHeight());
        mapRectFromViewToWindowCoords(view, mBoundingBox);

        outputBuffer[0] = Math.round(mBoundingBox.left);
        outputBuffer[1] = Math.round(mBoundingBox.top);
        outputBuffer[2] = Math.round(mBoundingBox.right - mBoundingBox.left);
        outputBuffer[3] = Math.round(mBoundingBox.bottom - mBoundingBox.top);
    }

    private void mapRectFromViewToWindowCoords(View view, RectF rect) {
        Matrix matrix = view.getMatrix();
        if (!matrix.isIdentity()) {
            matrix.mapRect(rect);
        }

        rect.offset(view.getLeft(), view.getTop());

        ViewParent parent = view.getParent();
        while (parent instanceof View) {
            View parentView = (View) parent;

            rect.offset(-parentView.getScrollX(), -parentView.getScrollY());

            matrix = parentView.getMatrix();
            if (!matrix.isIdentity()) {
                matrix.mapRect(rect);
            }

            rect.offset(parentView.getLeft(), parentView.getTop());

            parent = parentView.getParent();
        }
    }

    ViewHierarchy(View root, SparseArray viewMap) {
        super();
        this.root = root;
        this.viewMap = viewMap;
        this.RESOURCES_GUARD = new Semaphore(1);
    }


    public static final class Companion {
        public final String toXmlNodeName(String className) {
            if (className != null) {
                int startIndex$iv$iv = className.length() - 1;
                boolean startFound$iv$iv = false;

                boolean match$iv$iv;
                while (startIndex$iv$iv <= startIndex$iv$iv) {
                    int index$iv$iv = !startFound$iv$iv ? startIndex$iv$iv : startIndex$iv$iv;
                    char it = className.charAt(index$iv$iv);
                    match$iv$iv = it <= ' ';
                    if (!startFound$iv$iv) {
                        if (!match$iv$iv) {
                            startFound$iv$iv = true;
                        } else {
                            ++startIndex$iv$iv;
                        }
                    } else {
                        if (!match$iv$iv) {
                            break;
                        }

                        --startIndex$iv$iv;
                    }
                }

                CharSequence temp = className.subSequence(startIndex$iv$iv, startIndex$iv$iv + 1).toString();
                if (temp.length() != 0) {
                    String regex = "[$@#&]";
                    String replacement = ".";
                    className = className.replaceAll(regex, replacement);
                    regex = "\\.+";
                    replacement = ".";
                    className = className.replaceAll(regex, replacement);
                    regex = "(^\\.|\\.$)";
                    replacement = "";
                    className = className.replaceAll(regex, replacement);
                    String fixedName = className.replaceAll(regex, replacement);
                    fixedName = this.toNodeName(fixedName);
                    String $this$trim$iv$iv = fixedName;
                    startIndex$iv$iv = 0;
                    int endIndex$iv$iv = $this$trim$iv$iv.length() - 1;
                    startFound$iv$iv = false;

                    while (startIndex$iv$iv <= endIndex$iv$iv) {
                        int index$iv$iv = !startFound$iv$iv ? startIndex$iv$iv : endIndex$iv$iv;
                        char it = $this$trim$iv$iv.charAt(index$iv$iv);
                        match$iv$iv = it <= ' ';
                        if (!startFound$iv$iv) {
                            if (!match$iv$iv) {
                                startFound$iv$iv = true;
                            } else {
                                ++startIndex$iv$iv;
                            }
                        } else {
                            if (!match$iv$iv) {
                                break;
                            }

                            --endIndex$iv$iv;
                        }
                    }

                    temp = $this$trim$iv$iv.subSequence(startIndex$iv$iv, endIndex$iv$iv + 1).toString();
                    if (temp.length() == 0) {
                        fixedName = "android.view.View";
                    }

                    return fixedName;
                }
            }

            return "android.view.View";
        }

        private final String toNodeName(String str) {
            String nodeName = ViewHierarchy.XML10_START_TAG_PATTERN.matcher(str).replaceAll("");
            return ViewHierarchy.XML10_TAG_PATTERN.matcher(nodeName).replaceAll("_");
        }

        public final void $setMaxTraverseDepth(int maxTraverseDepth) {
            ViewHierarchy.MAX_TRAVERSAL_DEPTH = maxTraverseDepth;
        }

        private Companion() {
        }

    }
}
